
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <baytech_module_harness.h>
#include <baytech_slave_context.h>
#include <prototypes.h>
#include <pthread.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_baytech_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static void set_device_address(CONTEXT *ctx, char *dev);
static int baytech_engage(CONTEXT *ctx);
static void baytech_disengage(CONTEXT *ctx);
static int bt_printf (int fd, const char *fmt, ...);


/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
  baytech_disengage(ctx);
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			continue;
			}
		if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		pthread_testcancel();
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",
				 ctx->node_id,
				 msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_baytech_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;

		/* BEGIN San mod 07/12/00 */
		value = &saved_buffer[((variable - msg_buffer)+(strlen(variable)+1))];
		/* END San mod 07/12/00 */

		if (!strcasecmp(variable,"DEVICE_ADDRESS"))
			{
      baytech_disengage(ctx);
      set_device_address(ctx, value);
      if(strcasecmp(value,"NONE"))
        {
        if ((rc = baytech_engage(ctx)) < 0)
          {
          sprintf(log_msg,"0:0:LOG:Thread %s unable to engage on %s",
            ctx->node_id,ctx->device_address);
          send_msg(ctx->thread_ipc_write, log_msg);
          }
        }
      goto out;
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char								*fd;
		int									client_fd;

		/* Called when a client disconnects from nexxus */
		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
		goto out;
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	*new_name;
	
		/* Called when a node is renamed */
		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
    free(ctx->node_id);
    ctx->node_id = (char *) malloc(strlen(new_name)+1);
    strcpy(ctx->node_id, new_name);
		goto out;
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Process an IPC message
 */
static void process_baytech_ipc(CONTEXT			*ctx,
															__uint32_t	job_id,
															int					client_fd,
															char				*ipc_cmd,
															char				*ipc_args)
{
  char          log_msg[255];
	char					resp_msg[255];
	char					*last = (char *) 0x00;

	if (!strcasecmp(ipc_cmd, "CONFIGURATION"))
		{
    char *dev;
    CONTEXT *ctx_scan;

    if (!(dev = strtok_r(ipc_args, ":", &last)))
      goto out_malformatted;

    if (strcasecmp(dev,"NONE"))
      {
      pthread_mutex_lock(&context_mutex);
      ctx_scan = context_head;
      while(ctx_scan)
        {
        if (ctx_scan->device_address)
          if ((!strcmp(ctx_scan->device_address,dev)) && (ctx_scan != ctx))
            break;
        ctx_scan = ctx_scan->next;
        }
      pthread_mutex_unlock(&context_mutex);

      if (ctx_scan)
        {
        sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s", 
          job_id,client_fd,strerror(EADDRINUSE));
        send_msg(ctx->thread_ipc_write, resp_msg);
        goto out_ok;
        }

      /* Has the serial port changed? */
      if ((ctx->device_address ? strcmp(ctx->device_address, dev) : 1))
        {
        set_device_address(ctx, dev);
        baytech_disengage(ctx);

        /* Do we need to reengage */
        if (strcasecmp(dev, "NONE"))
          {
          int rc;
          if ((rc = baytech_engage(ctx) <0))
            {
            sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
              job_id,client_fd,strerror(EINVAL));
            send_msg(ctx->thread_ipc_write, resp_msg);
            goto out_ok;
            }
          }
        sprintf(resp_msg,"0:0:SET:%s:BAYTECH:DEVICE_ADDRESS:%s",
          ctx->node_id,dev);
        send_msg(ctx->thread_ipc_write, resp_msg);
        }
      sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",
        job_id,client_fd);
      send_msg(ctx->thread_ipc_write, resp_msg);
      goto out_ok;
      }
		}

  if (!ctx->device_address)
  {
    sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
    job_id,client_fd,strerror(ENODEV));
    send_msg(ctx->thread_ipc_write, resp_msg);
    goto out_ok;
    }

	if (!strcasecmp(ipc_cmd, "INQUIRY"))
		{
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:%s:%s",job_id,client_fd,
      ctx->node_id, ctx->device_address);
		send_msg(ctx->thread_ipc_write, resp_msg);
		}
	else if (!strcasecmp(ipc_cmd, "RESET"))
    {
    /* Reset the Baytech unit */
    bt_printf(ctx->con_sock, "\r");
    bt_printf(ctx->con_sock, "5\r");
    bt_printf(ctx->con_sock, "Y\r");
    baytech_disengage(ctx);
    sleep(5);
    baytech_engage(ctx);
    }
	else if (!strcasecmp(ipc_cmd, "PORT_ON"))
		{
		int port;
		char *bword;
		
    if (!(bword = strtok_r(ipc_args, ":", &last)))
      goto out_malformatted;
		port = atoi(bword);

		sprintf(log_msg,
				"0:0:LOG:Powering on port %d of unit %s",
		 	  port,ctx->node_id);
		send_msg(ctx->thread_ipc_write, log_msg);

    bt_printf(ctx->con_sock, "\r");
    bt_printf(ctx->con_sock, "1\r");

    bt_printf (ctx->con_sock, "ON %d\r", port);
    bt_printf (ctx->con_sock, "Y\r", port);
    bt_printf(ctx->con_sock, "MENU\r");
		}
	else if (!strcasecmp(ipc_cmd, "PORT_OFF"))
		{
		int port;
		char *bword;
		
    if (!(bword = strtok_r(ipc_args, ":", &last)))
      goto out_malformatted;
		port = atoi(bword);

		sprintf(log_msg,
				"0:0:LOG:Powering off port %d of unit %s",
		 	  port,ctx->node_id);
		send_msg(ctx->thread_ipc_write, log_msg);

    bt_printf(ctx->con_sock, "\r");
    bt_printf(ctx->con_sock, "1\r");

    bt_printf (ctx->con_sock, "OFF %d\r", port);
    bt_printf (ctx->con_sock, "Y\r");
    bt_printf(ctx->con_sock, "MENU\r");
		}
	else if (!strcasecmp(ipc_cmd, "PORT_REBOOT"))
		{
		int port;
		char *bword;
		
    if (!(bword = strtok_r(ipc_args, ":", &last)))
      goto out_malformatted;
		port = atoi(bword);

		sprintf(log_msg,
				"0:0:LOG:Rebooting port %d of unit %s",
		 	  port,ctx->node_id);
		send_msg(ctx->thread_ipc_write, log_msg);

    bt_printf(ctx->con_sock, "\r");
    bt_printf(ctx->con_sock, "1\r");

    bt_printf (ctx->con_sock, "REBOOT %d\r", port);
    bt_printf (ctx->con_sock, "Y\r");
    sleep(11);
    bt_printf(ctx->con_sock, "MENU\r");
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
    sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
    send_msg(ctx->thread_ipc_write, resp_msg);
    return;

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}
	
/*
 * Receive a message. This function is threadsafe
 */
static int recv_msg(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
	char				*offset;
  int         rc;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
	else if (!rc)
		return(-1);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int send_msg(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

	if (fd == -1)
		return(-EBADF);
  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}

static void set_device_address(CONTEXT *ctx, char *dev)
{
  if (ctx->device_address)
    free(ctx->device_address);
  ctx->device_address = malloc((strlen(dev)+1));
  strcpy(ctx->device_address, dev);
}

static int baytech_engage(CONTEXT *ctx)
{
  char *last = (char *) 0x00;
  char log_msg[255];
#ifndef __sgi
  struct serial_struct  ser;
#endif /* __sgi */

  if (ctx->con_sock != -1)
    return(-EADDRINUSE);

  if (!ctx->device_address)
    return(-EFAULT);

  /* Is this a serial port or a network address */
  if (ctx->device_address[0] == '/')
    {
    struct termios tty;
    if ((ctx->con_sock = open(ctx->device_address, O_RDWR))<0)
      {
      sprintf(log_msg,"0:0:LOG:Thread %s failed to engage on serial port (%m)",
        ctx->node_id);
      send_msg(ctx->thread_ipc_write, log_msg);
      return(-errno);
      }
    tcgetattr(ctx->con_sock, &tty);
    cfmakeraw(&tty);
    cfsetospeed(&tty,B9600);
    cfsetispeed(&tty,B9600);
    tty.c_cflag &= ~CRTSCTS;
    tty.c_lflag &= ~ICANON;
    tcsetattr(ctx->con_sock,TCSANOW,&tty);
#ifndef __sgi
    if (ioctl (ctx->con_sock, TIOCGSERIAL, &ser) == 0) 
      {
      if (!(ser.flags & ASYNC_LOW_LATENCY)) 
        {
        ser.flags |= ASYNC_LOW_LATENCY;
        ioctl (ctx->con_sock, TIOCSSERIAL, &ser);
        }
      }
#endif  /*__sgi*/
    }
  else
    {
    struct sockaddr_in sock;
    int optval;
    char address[64];
    unsigned int port;
    char *bword;
    char tmp[64];

    strcpy(tmp, ctx->device_address);

    if (!(bword = strtok_r(tmp, "-", &last)))
      return(-EFAULT);
    strcpy(address,bword);

    if (!(bword = strtok_r(NULL, "-", &last)))
      return(-EFAULT);

    port = atoi(bword);

    if ((ctx->con_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
      {
      sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
      send_msg(ctx->thread_ipc_write, log_msg);
      return(-errno);
      }
    optval = 1;
    if (setsockopt(ctx->con_sock,SOL_SOCKET,SO_REUSEADDR,
      &optval,sizeof(optval))<0)
      {
      sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
      send_msg(ctx->thread_ipc_write, log_msg);
      return(-errno);
      }

    memset(&sock, 0, sizeof(sock));
    sock.sin_family = AF_INET;
    sock.sin_addr.s_addr = inet_addr(address);
    sock.sin_port = htons(port);

    if (connect(ctx->con_sock,(struct sockaddr *) &sock,sizeof(sock)) < 0)
      {
      sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
      send_msg(ctx->thread_ipc_write, log_msg);
      return(-errno);
      }
    }
  sprintf(log_msg,"0:0:LOG:Thread %s engage successful",ctx->node_id);
  send_msg(ctx->thread_ipc_write, log_msg);
  return(0);
}

/*   Disengage from the baytech */
static void baytech_disengage(CONTEXT *ctx)
{
  if (ctx->con_sock != -1)
    {
    char log_msg[255];
    int tmp_sock;

    sprintf(log_msg, "0:0:LOG:Thread %s disengaging from %s",
      ctx->node_id, ctx->device_address);
    send_msg(ctx->thread_ipc_write, log_msg);

    tmp_sock = ctx->con_sock;
    ctx->con_sock = -1;
    close(tmp_sock);
    }
}

static int bt_printf (int fd, const char *fmt, ...)
{
  va_list ap;
  char *buf;
  int n;

  buf = (char *)malloc (1024);
  memset (buf, 0, 1024);

  va_start (ap, fmt);
  n = vsnprintf (buf, 1024, fmt, ap);
  va_end (ap);

  if (n <= 0) 
    {
    free(buf);
    return(-1);
    }
  else
    n = write (fd, buf, strlen (buf));

  free(buf);
  sleep(1);
  return(n);
}
