/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <libloose.h>
#include <rsh_module_harness.h>
#include <rsh_slave_context.h>
#include <prototypes.h>
#include <pthread.h>

CONTEXT	*context_head = NULL;
CONTEXT	*context_tail = NULL;

static void			context_free(CONTEXT *context);
static CONTEXT	*context_alloc(char *node_id);
static void 		context_free_resources(CONTEXT *ctx);
static int 			thread_send(CONTEXT *ctx, char *msg);

/*
 * Now and then a thread may want to peek at another threads
 * context structure. This lock is in place to keep bad things
 * from happening
 */
pthread_mutex_t	context_mutex;

void context_init()
{
	pthread_mutex_t tmp_mutex = PTHREAD_MUTEX_INITIALIZER;

	memcpy(&context_mutex, &tmp_mutex, sizeof(pthread_mutex_t));
	context_head = NULL;
	context_tail = NULL;
	pthread_mutex_init(&context_mutex, NULL);
}

/*
 * Set a context variable and notify the thread
 */
int context_variable_set(char *node_id, char *variable, char *value)
{
	CONTEXT *ctx;
	char		thread_msg[255];

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	sprintf(thread_msg,"NODE_VAR:%s:%s:%s",node_id,variable,value);
	return(thread_send(ctx, thread_msg));
}

/*
 * Rename a node and notify the thread
 */
int context_rename(char *node_id, char *new_id)
{
	CONTEXT *ctx;
	char		thread_msg[255];

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	sprintf(thread_msg,"NODE_RENAME:%s:%s",node_id,new_id);
	return(thread_send(ctx, thread_msg));
}

/*
 * Send IPC to a node thread
 */
int context_send_ipc(char *node_id, char *ipc_msg)
{
	CONTEXT *ctx;

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	return(thread_send(ctx, ipc_msg));
}

/*
 * Cleanly destroy a node context
 */
int context_destroy(char *node_id)
{
	CONTEXT					*ctx;

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	if (pthread_cancel(ctx->thread) < 0)
		return(-EBUSY);
	pthread_join(ctx->thread, NULL);
	context_free(ctx);
	return(0);
}

/*
 * Notify all threads about a clients disconnection
 */
void context_notify_disconnection_all(int fd)
{
	CONTEXT *ctx;
	char		msg[255];

	sprintf(msg,"CLIENT_DIS:NONE:%d",fd);
	ctx = context_head;
	while(ctx)
		{
		thread_send(ctx,msg);
		ctx = ctx -> next;
		}
}

/*
 * Create a context for a node and start it's thread
 */
CONTEXT *context_create(char *node_id)
{
	CONTEXT	*ctx;
	int			pipefd1[2], pipefd2[2];

	if (!(ctx = context_alloc(node_id)))
		return(NULL);
	if ((pipe(pipefd1) < 0) || (pipe(pipefd2) < 0))
		{
		context_free(ctx);
		return(NULL);
		}
	ctx->thread_ipc_read = pipefd2[0];
	ctx->thread_ipc_write= pipefd1[1];
	ctx->slave_ipc_read = pipefd1[0];
	ctx->slave_ipc_write= pipefd2[1];

	if (pthread_create(&ctx->thread, NULL, thread_start, (void *) ctx) < 0)
		{
		printf("RSH Slave pthread_create() call failed for node %s\n",
					 node_id);
		context_free(ctx);
		return(NULL);
		}
	return(ctx);
}

void context_destroy_all()
{
	CONTEXT *ctx;
	int			rc;

	ctx = context_head;
	while(ctx)
		{
		if ((rc = context_destroy(ctx->node_id)) < 0)
			{
			printf("Slave unable to destroy thread for node %s (%s)\n",
						 ctx->node_id,
						 strerror(-rc));
			}
		// ctx = ctx -> next;
		ctx = context_head;
		}
}

CONTEXT *context_lookup(char *node_id)
{
	CONTEXT	*ctx;

	ctx = context_head;
	while(ctx)
		{
		if (!strcasecmp(node_id, ctx->node_id))
			return(ctx);
		ctx = ctx -> next;
		}
	return(NULL);
}

/*
 * Free any remaining resources for a context
 * This is called *after* the thread is cancelled
 */
static void context_free_resources(CONTEXT *ctx)
{
	close(ctx->thread_ipc_read);
	close(ctx->thread_ipc_write);
	close(ctx->slave_ipc_read);
	close(ctx->slave_ipc_write);
	ctx->thread_ipc_read = -1;
	ctx->thread_ipc_write = -1;
	ctx->slave_ipc_read = -1;
	ctx->slave_ipc_write = -1;
	if (ctx->node_id)
		{
		free(ctx->node_id);
		ctx->node_id = NULL;
		}
	if (ctx->agent)
		{
		free(ctx->agent);
		ctx->agent = NULL;
		}
	if (ctx->addr)
		{
		free(ctx->addr);
		ctx->addr = NULL;
		}
	if (ctx->user)
		{
		free(ctx->user);
		ctx->addr = NULL;
		}
}

static CONTEXT *context_alloc(char *node_id)
{
	CONTEXT	*new;

	if (!node_id)
		{
		lm_log("NULL passed in as node_id for new context");
		return(NULL);
		}
	if (!(new = (CONTEXT *) malloc(sizeof(CONTEXT))))
		{
		lm_log("Unable to malloc new context structure");
		return(NULL);
		}
	memset(new, 0, sizeof(CONTEXT));
	
	pthread_mutex_lock(&context_mutex);

	new->thread_ipc_read = -1;
	new->thread_ipc_write = -1;
	new->slave_ipc_read = -1;
	new->slave_ipc_write = -1;
	new->con_sock = -1;
  if (!context_head)
    {
    context_head = new;
    context_tail = new;
    }
  else
    {
    context_tail->next = new;
    new->prev = context_tail;
    context_tail = new;
    }
	new->node_id = malloc((strlen(node_id)+1));
  strcpy(new->node_id, node_id);

	pthread_mutex_unlock(&context_mutex);

	return(new);
}

static void context_free(CONTEXT *context)
{
	pthread_mutex_lock(&context_mutex);

  if (context == context_head)
    {
    context_head = context_head->next;
    if (context_head)
      context_head->prev = NULL;
    }
  else if (context == context_tail)
    {
    context_tail = context_tail->prev;
    if (context_tail)
      context_tail->next = NULL;
    }
  else
    {
    context->next->prev = context->prev;
    context->prev->next = context->next;
    }
	context_free_resources(context);

	pthread_mutex_unlock(&context_mutex);

	free(context);
}

static int thread_send(CONTEXT *ctx, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (ctx->slave_ipc_write,
									 (void *) &net_len,
									 sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (ctx->slave_ipc_write,
									 (void *) msg,
									 len)) < 0)
    return(rc);
  return(0);
}

