#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <ctype.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <libloose.h>
#include <prototypes.h>

#define VASENET_MAJOR_VERSION 1
#define VASENET_MINOR_VERSION 1

#define VASENET_PORT 53005

#define VASENET_CONNECT_TIMEOUT 5
#define VASENET_READ_TIMEOUT 4
#define VASENET_RECONNECT_TIMER 120
#define VASENET_PING_TIMER 20

#define NO_ERROR 0

#define MASTER_NODE 1
#define SLAVE_NODE 2

#define KNOWN_UNUSED_PARAM(x) { int __bleh; __bleh = (int) x;}

typedef struct nodelist
	{
	struct 			nodelist *next;
	struct 			nodelist *prev;

	char 				*node_id;
	int 				type;

	/* VA 100 */
	char 				ip_address[17];
	char 				*password;
	int 				con_sock;

	/* VA 1220 */
	char 				vasenet_address[7];
	char 				*master;

	} NODELIST;

static void config_cb(__uint32_t job_id, int client_fd, char *msg);
static void deinit_cb(__uint32_t job_id, int client_fd, char *msg);
static void discon_cb(__uint32_t job_id, int client_fd, char *msg);
static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg);

static int connect_timer(void *arg);
static int ping_timer(void *arg);
void log_printf(char *fmt, ...);
void send_final_response(int err, __uint32_t job_id, __uint32_t client_fd);

static int vasenet_engage(NODELIST *node);
static void vasenet_disengage(NODELIST *node);
static int vasenet_authenticate(NODELIST *node);
static int vasenet_query(NODELIST *node, char *cmd, char **response);
static char vasenet_read_char(NODELIST *node);
static void vasenet_send_error(int errnum, __uint32_t job_id, __uint32_t  client_fd);
static void vasenet_update_va100_display(NODELIST *node);
static int vasenet_is_configured_master(NODELIST *node);

int do_configuration(NODELIST *node);
int do_inquiry(NODELIST *, __uint32_t, __uint32_t);
int do_vasenet_version(NODELIST *, __uint32_t, __uint32_t);
int do_node_list(NODELIST *, __uint32_t, __uint32_t);
int do_refresh(NODELIST *, __uint32_t, __uint32_t);
int do_status(NODELIST *, __uint32_t, __uint32_t);
int do_reset(NODELIST *, __uint32_t, __uint32_t);
int do_rescan(NODELIST *, __uint32_t, __uint32_t);
int do_identify(NODELIST *, __uint32_t, __uint32_t);
int do_power_on(NODELIST *, __uint32_t, __uint32_t);
int do_power_off(NODELIST *, __uint32_t, __uint32_t);
int do_hard_reset(NODELIST *, __uint32_t, __uint32_t);
int do_chassis_status(NODELIST *, __uint32_t, __uint32_t);

/* Linked List Code */

static NODELIST *nodelist_head = NULL;
static NODELIST *nodelist_tail = NULL;

static NODELIST *node_create(char *node_id)
{
	NODELIST *new;

	if (!(new = (NODELIST *) malloc(sizeof(NODELIST))))
		return(0);

	memset(new, 0, sizeof(NODELIST));

	if (!nodelist_head)
		nodelist_head = nodelist_tail = new;
	else
		{
		nodelist_tail->next = new;
		new->prev = nodelist_tail;
		nodelist_tail = new;
		}

	/* fill in the node structure */
	new->node_id = strdup(node_id);
	new->con_sock = -1;
	new->ip_address[0] = '\0';
	new->password = NULL;
	
	new->vasenet_address[0] = '\0';
	new->master = NULL;

	return(new);
}

static NODELIST *lookup_node(char *node_id)
{
	NODELIST *scan;

	if (!node_id)
		return(NULL);
	scan = nodelist_head;
	while(scan)
		{
		if (!strcasecmp(node_id, scan->node_id))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

static void node_destroy(NODELIST *node)
{
	if (node == nodelist_head)
		{
		nodelist_head = nodelist_head->next;
		if (nodelist_head)
			nodelist_head->prev = NULL;
		else
			nodelist_tail = NULL;
		}
	else if (node == nodelist_tail)
		{
		nodelist_tail = nodelist_tail->prev;
		if (nodelist_tail)
			nodelist_tail->next = NULL;
		else
			nodelist_head = NULL;
		}
	else
		{
		node->next->prev = node->prev;
		node->prev->next = node->next;
		}

	if (node->node_id)
		free (node->node_id);

	if (node->password)
		free (node->password);

	if (node->master)
		free (node->master);
	free(node);
}

/* Module Code */

int main(int argc, char **argv)
{

	KNOWN_UNUSED_PARAM(argc);
	KNOWN_UNUSED_PARAM(argv);

	lm_init();
	lm_nexxus_connect();
	if (lm_register(	"VASENET",
													"VASENET Administration Module",
													"Module to support VA 100 based clusters",
													"Zac Sprackett <zacs@valinux.com>",
													"VASENET",
													VASENET_MAJOR_VERSION,
													VASENET_MINOR_VERSION) < 0) 
		{
		printf("Unable to register Module with Nexxus.\n");
		exit(-1);
		}
	
	lm_timer_add(VASENET_RECONNECT_TIMER,connect_timer,NULL);
	lm_timer_add(VASENET_PING_TIMER,ping_timer,NULL);

	lm_main_loop(config_cb,deinit_cb,discon_cb,ipc_cb);

	exit(0);
}

static void config_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST *node;
	char *cmd_module;
	char *node_id;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	cmd_module = strtok(msg, ":");
	node_id = strtok(NULL, ":");

	node = lookup_node(node_id);

	if (!strcasecmp(cmd_module, "NODE"))
		{
		if (node != NULL)
			log_printf("Received node config, but node %s already configured.",
				node_id);
		else
			node = node_create(node_id);
		}
	else if (!strcasecmp(cmd_module, "END_NODE"))
		{
		/* Do nothing */
		}
	else if (!strcasecmp(cmd_module, "ADDITION"))
		{
		if (node != NULL)
			log_printf("Received node addition, but node %s already exists.",
				node_id);
		else
			node = node_create(node_id);
		}
	else if (!strcasecmp(cmd_module, "DELETION"))
		{
		if (node)
			{
			NODELIST *scan;

			if (node->type == MASTER_NODE)
				{
				scan = nodelist_head;
				while(scan)
					{
					if (!(strcasecmp(node->node_id,scan->master)))
						{
						free(scan->master);
						scan->master = NULL;
						}
					scan = scan->next;
					}
				}
			node_destroy(node);
			}
		else
			log_printf("Received deletion event, but node %s doesn't exist.",
				node_id);
		}
	else if (!strcasecmp(cmd_module, "RENAME"))
		{
		char *new_name;
		char *old_name;

		if (!(new_name = strtok(NULL,":")))
			{
			log_printf("Received malformatted rename event for node %s.",
				node_id);
			return;
			}
		old_name = node->node_id;
		node->node_id = strdup(new_name);

		if (node->type == SLAVE_NODE) 
			{
			if (node->master)
				vasenet_update_va100_display(node);
			}
		else
			{
			NODELIST *scan;

			scan = nodelist_head;
			while (scan)
				{
				if ((scan->type == SLAVE_NODE) && (scan->master) && 
					(!(strcasecmp(scan->master, old_name))))
						{
						char resp_msg[255];

						free(scan->master);
						scan->master = strdup(new_name);

						sprintf(resp_msg,
							"0:0:SET:%s:VASENET:MASTER:%s",scan->node_id,scan->master);
						lm_send_to_nexxus(resp_msg);
						}
				scan = scan->next;
				}
			}
		free(old_name);
		}
	else if (!strcasecmp(cmd_module, "GLOBAL"))
		{
		char *variable, *value;

		if ((!(variable = strtok(NULL,":"))) || (!(value = strtok(NULL,":"))))
			{
			log_printf("Received malformatted local config message.");
			return;
			}

		if (!strcasecmp(variable, "IP_ADDRESS"))
			{
			snprintf(node->ip_address,sizeof(node->ip_address),"%s",value);

			if (!vasenet_is_configured_master(node))
				{
				vasenet_disengage(node);
				vasenet_engage(node);
				}
			}
		else
			log_printf("Received unknown global config event %s.",variable);
		}
	else if (!strcasecmp(cmd_module, "VASENET"))
		{
		char *variable, *value;

		/* Process our local variables */
		if ((!(variable = strtok(NULL,":"))) || (!(value = strtok(NULL,":"))))
			{
			log_printf("Received malformatted local config message.");
			return;
			}

		if (!strcasecmp(variable, "NODE_TYPE"))
			{
			int type = atoi(value);

			if ((type == MASTER_NODE) || (type == SLAVE_NODE))
				node->type = type;
			else
				node->type = 0;

			if (node->type == MASTER_NODE)
				{
				if (!vasenet_is_configured_master(node))
					{
					vasenet_disengage(node);
					vasenet_engage(node);
					}
				}
			}
		else if (!strcasecmp(variable, "PASSWORD"))
			{
			if (node->password)
				free(node->password);
			node->password = strdup(value);

			if (!vasenet_is_configured_master(node))
				{
				vasenet_disengage(node);
				vasenet_engage(node);
				}
			}
		else if (!strcasecmp(variable, "MASTER"))
			{
			if (node->master)
				free(node->master);
			node->master = strdup(value);
			}
		else if (!strcasecmp(variable, "VASENET_ADDRESS"))
			{
			snprintf(node->vasenet_address,sizeof(node->vasenet_address),"%s",value);
			}
		else
			log_printf("Received unknown local config event %s.",variable);
		}
	else
		log_printf("Received unknown config event %s.",cmd_module);
}

static void discon_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);
}

static void deinit_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);

	lm_nexxus_disconnect();
	exit(0);
}

static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg)
{
	NODELIST 		*node;
	char 				*cmd;
	char				*scratch;
	int 				error = NO_ERROR;

	if (!(cmd = strtok(msg, ":")) || !(scratch = strtok(NULL, ":")))
		{
		send_final_response(EINVAL,job_id,client_fd);
		return;
		}

	node = lookup_node(node_id);

	if (!node)
		{
		log_printf("Failed to look up node %s.",node_id);
		send_final_response(EDESTADDRREQ,job_id,client_fd);
		return;
		}


	/* All commands except configuration and inquiry require configured node */
	if ((node->type != 1) && (node->type != 2) &&
		(strcasecmp(cmd, "CONFIGURATION") !=0) && (strcasecmp(cmd,"INQUIRY") !=0))
			{
			send_final_response(EDESTADDRREQ,job_id,client_fd);
			return;
			}

	if (!strcasecmp(cmd, "CONFIGURATION"))
		{
		error = do_configuration(node);
		}
	else if (!strcasecmp(cmd, "INQUIRY"))
		{
		error = do_inquiry(node, job_id, client_fd);
		}
	else if (!strcasecmp(cmd, "VASENET_VERSION"))
		{
		error = do_vasenet_version(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "NODE_LIST")) && (node->type == MASTER_NODE))
		{
		error = do_node_list(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "REFRESH")) && (node->type == MASTER_NODE))
		{
		error = do_refresh(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "STATUS")) && (node->type == MASTER_NODE))
		{
		error = do_status(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "RESET")) && (node->type == MASTER_NODE))
		{
		error = do_reset(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "RESCAN")) && (node->type == MASTER_NODE))
		{
		error = do_rescan(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "IDENTIFY")) && (node->type == SLAVE_NODE))
		{
		error = do_identify(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "POWER_ON")) && (node->type == SLAVE_NODE))
		{
		error = do_power_on(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "POWER_OFF")) && (node->type == SLAVE_NODE))
		{
		error = do_power_off(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "HARD_RESET")) && (node->type == SLAVE_NODE))
		{
		error = do_hard_reset(node, job_id, client_fd);
		}
	else if ((!strcasecmp(cmd, "CHASSIS_STATUS")) && (node->type == SLAVE_NODE))
		{
		error = do_chassis_status(node, job_id, client_fd);
		}
	else
		error = EBADRQC;

	send_final_response(error, job_id, client_fd);
	return;
}

static int connect_timer(void *arg) {
	NODELIST *scan;

	KNOWN_UNUSED_PARAM(arg);

	scan = nodelist_head;
	while(scan)
		{
		if ((!(vasenet_is_configured_master(scan))) && (scan->con_sock == -1))
			vasenet_engage(scan);

		scan = scan->next;
		}

	return(1);
}

static int ping_timer(void *arg) {
	NODELIST *scan;

	KNOWN_UNUSED_PARAM(arg);

	scan = nodelist_head;
	while(scan)
		{
		if ((!(vasenet_is_configured_master(scan))) && (scan->con_sock != -1))
			{
			char *response = NULL;

			vasenet_query(scan,"S000000",&response);
			if (response)
				free(response);
			}
		scan = scan->next;
		}

	return(1);
}

int do_configuration(NODELIST *node)
{
	char resp_msg[1024];
	char *str;
	int  type;

	if (!(str = strtok(NULL, ":")))
		return(EINVAL);

	if (!strcasecmp(str,"MASTER"))
		type = MASTER_NODE;
	else if (!strcasecmp(str,"SLAVE"))
		type = SLAVE_NODE;
	else
		return(EINVAL);

	node->type = type;

	if (type == MASTER_NODE)
		{
		char *password;

		if ((!(password = strtok(NULL,":"))) || (strstr(password,":")) ||
				(strlen(password) != 8))
			return(EINVAL);

		if ((!node->password) || (strcmp(node->password,password)))
			{
			sprintf(resp_msg,
				"0:0:SET:%s:VASENET:PASSWORD:%s",node->node_id,password);
			lm_send_to_nexxus(resp_msg);

			if (node->password)
				free (node->password);
			node->password = strdup(password);

			vasenet_disengage(node);
			vasenet_engage(node);
			}
		else
			{
			if (node->con_sock == -1)
				vasenet_engage(node);
			}
		}
	else
		{
		char *master;
		char *addy;
		int i;

		if ((!(master = strtok(NULL,":"))) || (!(addy = strtok(NULL,":"))))
			return(EINVAL);

		if (strlen(addy) != 6)
			return(EINVAL);

		for (i = 0; i < (int)strlen(addy); i++)
			{
			if (!isxdigit(addy[i]))
				return(EINVAL);
			}

		sprintf(resp_msg,
			"0:0:SET:%s:VASENET:MASTER:%s",node->node_id,master);
		lm_send_to_nexxus(resp_msg);

		sprintf(resp_msg,
			"0:0:SET:%s:VASENET:VASENET_ADDRESS:%s",node->node_id,addy);
		lm_send_to_nexxus(resp_msg);

		if (node->master)
			free (node->master);
		node->master = strdup(master);
		snprintf(node->vasenet_address, sizeof(node->vasenet_address),"%s",addy);
		}

	sprintf(resp_msg,
		"0:0:SET:%s:VASENET:NODE_TYPE:%d",node->node_id,type);
	lm_send_to_nexxus(resp_msg);
	return(NO_ERROR);
}

int do_inquiry(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	char resp_msg[255];

	if (node->type == MASTER_NODE)
		{
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:MASTER:%s",
			job_id,client_fd,node->password);
		}
	else if ((node->type == SLAVE_NODE) && (node->master) && 
		(node->vasenet_address))
		{
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:SLAVE:%s:%s",
			job_id,client_fd,node->master,node->vasenet_address);
		}
	else
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:NOT_CONFIGURED",job_id,client_fd);

	lm_send_to_nexxus(resp_msg);

	return(NO_ERROR);
}

int do_vasenet_version(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	int rc = NO_ERROR;
	char *response = NULL;
	char resp_msg[255];

	if (node->type == MASTER_NODE)
		rc = vasenet_query(node,"V000000",&response);
	else
		{
		char cmd[32];

		snprintf(cmd,sizeof(cmd),"v%s",node->vasenet_address);
		rc = vasenet_query(node,"V000000",&response);
		}

	if (response)
		{
		sprintf(resp_msg,"%d:%d:FOR:VASENET_VERSION:%s",job_id,client_fd,response);
		lm_send_to_nexxus(resp_msg);
		free (response);
		}
	return(rc);
}

int do_node_list(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	int rc = NO_ERROR;
	char *response = NULL;

	rc = vasenet_query(node,"S000000",&response);

	if (!rc)
		{
		int num_nodes, i;
		char resp_msg[255];

		num_nodes = atoi(strtok(response, " "));
		free(response);

		for (i = 0; i<num_nodes; i++)
			{
			char cmd[64];

			sprintf(cmd,"g000000na%02d",i);
			rc = vasenet_query(node,cmd,&response);

			if (!rc)
				{
				sprintf(resp_msg,"%d:%d:FOR:NODE_LIST:%s",job_id, client_fd, response);
				lm_send_to_nexxus(resp_msg);
				}
			}
		}

	if (response)
		free(response);

	return(rc);
}

int do_refresh(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	vasenet_disengage(node);
	return(vasenet_engage(node));
}

int do_status(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	int rc = NO_ERROR;
	char *response = NULL;

	rc = vasenet_query(node,"S000000",&response);

	if (!rc)
		{
		char *nodes, *alarms;
		char resp_msg[255];

		nodes = strtok(response, " ");
		alarms = strtok(NULL, " ");
		sprintf(resp_msg,"%d:%d:FOR:STATUS:%s:%s",job_id, client_fd, nodes,alarms);
		lm_send_to_nexxus(resp_msg);
		}
	if (response)
		free(response);

	return(rc);
}

int do_reset(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{	
	int rc = NO_ERROR;
	char *response = NULL;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	rc = vasenet_query(node,"R000000",&response);
	if (response)
		free(response);

	if ((!rc) && (node->type == MASTER_NODE))
		vasenet_disengage(node);
	return(rc);
}

int do_rescan(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	int rc = NO_ERROR;
	char *response = NULL;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	rc = vasenet_query(node, "P000000", &response);
	if (response)
		free(response);

	sleep(3);

	if (!rc)
		vasenet_update_va100_display(node);

	return(rc);
}

int do_identify(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	char *response = NULL;
	char cmd[64];
	int rc = NO_ERROR;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	snprintf(cmd,sizeof(cmd),"!%s",node->vasenet_address);
	rc = vasenet_query(node,cmd,&response);

	if (response)
		free(response);

	return(rc);
}

int do_power_on(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{	
	char *response = NULL;
	char cmd[64];
	int rc = NO_ERROR;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	snprintf(cmd,sizeof(cmd),"u%s",node->vasenet_address);
	rc = vasenet_query(node,cmd,&response);

	if (response)
		free(response);

	return(rc);
}

int do_power_off(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	char *response = NULL;
	char cmd[64];
	int rc = NO_ERROR;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	snprintf(cmd,sizeof(cmd),"d%s",node->vasenet_address);
	rc = vasenet_query(node,cmd,&response);

	if (response)
		free(response);

	return(rc);
}

int do_hard_reset(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	char *response = NULL;
	char cmd[64];
	int rc = NO_ERROR;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	snprintf(cmd,sizeof(cmd),"r%s",node->vasenet_address);
	rc = vasenet_query(node,cmd,&response);

	if (response)
		free(response);

	return(rc);
}

int do_chassis_status(NODELIST *node, __uint32_t job_id, __uint32_t client_fd)
{
	char *response = NULL;
	char cmd[64];
	int rc = NO_ERROR;

	snprintf(cmd,sizeof(cmd),"s%s",node->vasenet_address);
	rc = vasenet_query(node,cmd,&response);

	if (!rc)
		{
		char resp_msg[255];
		sprintf(resp_msg,"%d:%d:FOR:CHASSIS_STATUS:%c:%c:%c:%c",
			job_id,client_fd,response[0],response[2],response[4],response[6]);
		lm_send_to_nexxus(resp_msg);
		}
	if (response)
		free(response);

	return(rc);
}

void log_printf(char *fmt, ...)
{
	char msg[1024];
	va_list ap;

	va_start(ap,fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);
	lm_log(msg);
}

void send_final_response(int err, __uint32_t job_id, __uint32_t client_fd)
{
	char resp[128];

	if (!err)
		{
		snprintf(resp, sizeof(resp), "%d:%d:FOR:JOB_COMPLETED", job_id, client_fd);
		lm_send_to_nexxus(resp);
		}
	else if (err < 0)
		vasenet_send_error(-err,job_id,client_fd);
	else
		{
		snprintf(resp, sizeof(resp), "%d:%d:FOR:JOB_ERROR:%s",
			job_id,client_fd, strerror(err));
		lm_send_to_nexxus(resp);
		}
}

/* vasenet base commands */

/* vasenet_engage() - returns 0 on success, errno on failure */
static int vasenet_engage(NODELIST *node)
{
	int optval, sock_opts, rc;
	struct sockaddr_in sock;
	struct timeval to;
	fd_set write_fds;

	if (node->type != MASTER_NODE)
		{
		log_printf("Tried to engage node %s which is not a master.",node->node_id);
		return(-1);
		}

	if ((!node->password) || (!node->ip_address))
		return(-1);

	if (node->con_sock != -1)
		{
		/* We are already engaged */
		return(0);
		}

	if ((node->con_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
		{
		log_printf("Thread %s engage failed (%m)",node->node_id);
		return(errno);
		}

	optval = 1;
	if (setsockopt(node->con_sock,
		SOL_SOCKET,
		SO_REUSEADDR,
		&optval,
		sizeof(optval)) < 0)
			{
			log_printf("Thread %s engage failed (%m)",node->node_id);
			vasenet_disengage(node);
			return(errno);
			}

	if ((sock_opts = fcntl(node->con_sock, F_GETFL))<0)
		{
		log_printf("Thread %s engage failed to get socket options (%m)",
			node->node_id);
		vasenet_disengage(node);
		return(errno);
		}

	/* Set the socket non blocking */
	if (fcntl(node->con_sock, F_SETFL, O_NONBLOCK)<0)
		{
		log_printf("Thread %s engage failed to set socket O_NONBLOCK (%m)",
			node->node_id);
		vasenet_disengage(node);
		return(-errno);
		}

	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = inet_addr(node->ip_address);
	sock.sin_port = htons(VASENET_PORT);

	if (connect(node->con_sock,(struct sockaddr *) &sock, sizeof(sock)) < 0)
		{
		if (errno != EINPROGRESS)
			{
			log_printf("Thread %s engage failed in connect (%m)",
				node->node_id);
			vasenet_disengage(node);
			return(errno);
			}
		}

	/* Select so we can set a time limit for connect() */
	FD_ZERO(&write_fds);
	FD_SET(node->con_sock,&write_fds);
	to.tv_sec = VASENET_CONNECT_TIMEOUT;
	to.tv_usec = 0;

	if ((rc = select((node->con_sock+1), NULL, &write_fds, NULL, &to)) < 1)
		{
		log_printf("Thread %s engage timed out in connect.",node->node_id);
		vasenet_disengage(node);
		return(ETIMEDOUT);
		}
	
	/* Set the socket to blocking again */
	if (fcntl(node->con_sock, F_SETFL, sock_opts)<0)
		{
		log_printf("Thread %s engage, failed to reset socket (%m)",
			node->node_id);
		vasenet_disengage(node);
		return(errno);
		}

	if ((rc = vasenet_authenticate(node)))
		{
		vasenet_disengage(node);
		return(rc);
		}

	vasenet_update_va100_display(node);
	return(0);
}

/* vasenet_update_va100_display() - call with a slave node to just update its
                                    name.  master name updates all its slaves. 
                                    NULL updates all nodes */
static void vasenet_update_va100_display(NODELIST *node)
{
	char cmd[64];
	char *response;

	if (node)
		{
		if (node->type == MASTER_NODE)
			{
			NODELIST *scan;

			scan = nodelist_head;
	
			while(scan)
				{
				if ((scan->type == SLAVE_NODE) && (scan->master) && 
					(scan->vasenet_address) && 
					(!(strcasecmp(scan->master,node->node_id))))

				vasenet_update_va100_display(scan);
				scan = scan->next;
				}
			return;
			}
		else
			{
			int rc; 

			snprintf(cmd,sizeof(cmd),"N%s%s",node->vasenet_address,node->node_id);
			rc = vasenet_query(node,cmd,&response);

			if (response)
				free (response);
			return;
			}
		}
	else
		{
		NODELIST *scan;

		scan = nodelist_head;

		while(scan)
			{
			vasenet_update_va100_display(scan);
			scan = scan->next;
			}
		}
}

static int vasenet_authenticate(NODELIST *node)
{
	char auth_key[17], write_buff[32], md5_buff[64];
	unsigned int chksum[4] = {0,0,0,0};
	int i;
	char *response;

	memset(auth_key,0,sizeof(auth_key));
	memset(md5_buff,0,sizeof(md5_buff));
	memset(write_buff,0,sizeof(write_buff));

	/* We need to select here to prevent reads from blocking forever */
	for (i = 0; i<16; i++)
		{
		if (!(auth_key[i] = vasenet_read_char(node)))
			return(ECONNRESET);
		}
		auth_key[16] = '\0';

	sprintf(md5_buff,"%s%s",auth_key,node->password);
	MD5Transform(chksum,(unsigned int *)md5_buff);
	sprintf(write_buff,"%08X%08X%08X%08X",
		chksum[0],chksum[1],chksum[2],chksum[3]);
	write(node->con_sock,write_buff,sizeof(write_buff));

	/* Send a command to verify our connection */
	if (vasenet_query(node,"S000000",&response))
		{
		log_printf("Failed to properly authenticate with node %s",node->node_id);
		return(EBADE);
		}

	if (response)
		free(response);

	return(0);
}

static void vasenet_disengage(NODELIST *node)
{
	if (node->con_sock != -1)
		close(node->con_sock);
	node->con_sock = -1;
}

/* vasenet_query() - Returns 0 on success... negative with VA100 passthrough
                     error, or positive errno. */
static int vasenet_query(NODELIST *node, 
	char *cmd,
	char **response)
{
	NODELIST *master_node;
	char query_buffer[128], length_buffer[3];
	char *index = NULL;
	int size, length,error_code, i, rc;

	*response = NULL;

	if (node->type == MASTER_NODE)
		master_node = node;
	else
		{
		if (!node->master)
			return(EBADR);

		if (!(master_node = lookup_node(node->master)))
			return(EBADR);
	
		if (master_node->type != MASTER_NODE)
			return(EBADR);
		}	

	if (!(strlen(master_node->ip_address)))
		return(EDESTADDRREQ);

	if (master_node->con_sock == -1)
			return(ENOTCONN);

	/* send our command */
	size = strlen(cmd) + 2;
	snprintf(query_buffer,sizeof(query_buffer),"%02d%s",size,cmd);
	rc = write(master_node->con_sock,query_buffer,size);

	if (rc != size)
		{
		log_printf("Attempted to write %d byte%s, wrote %d.",
			size,(size == 1 ?"":"s"),rc);
		vasenet_disengage(master_node);
		return(ECONNRESET);
		}

	memset(length_buffer,0,sizeof(length_buffer));
	for (i = 0; i < (int)sizeof(length_buffer); i++)
		{
		if (!(length_buffer[i] = vasenet_read_char(master_node)))
			{
			log_printf("Expected %d length bytes from VA100 %s, received %d.",sizeof(length_buffer),master_node->node_id,i);
			vasenet_disengage(master_node);
			return(ECONNRESET);
			}
		}
	length_buffer[2] = '\0';
	length = atoi(length_buffer) - 3;

	memset(query_buffer,0,sizeof(query_buffer));
	for(i = 0; i<length; i++)
		{
		if (!(query_buffer[i] = vasenet_read_char(master_node)))
			{
			log_printf("Failed to read response from VA100 %s, read %d, length %d.",
				master_node->node_id,(i-1),length);
			vasenet_disengage(master_node);
			return(ECONNRESET);
			}
		}

	if (strlen(query_buffer)>2)
		index = &query_buffer[2];

	query_buffer[1] = '\0';
	error_code = atoi(query_buffer);

	if (index)
		*response = strdup(index);

	/* VA100 errors return as negative numbers */
	if (error_code)
		return(-error_code);
	else
		return(0);
}

static char vasenet_read_char(NODELIST *node)
{
	struct timeval to;
	fd_set read_fds;
	int rc;
	char char_buff;

	FD_ZERO(&read_fds);
	to.tv_sec = VASENET_READ_TIMEOUT;
	to.tv_usec = 0;
	FD_SET(node->con_sock, &read_fds);

	if ((rc = select((node->con_sock+1),&read_fds,NULL,NULL,&to))<=0)
		{
		log_printf("select() timed out in vasenet_read_char, timeout was %d",
			to.tv_sec);
		vasenet_disengage(node);
		return((char)NULL);
		}

	if ((read(node->con_sock,&char_buff,1))<1)
		{
		vasenet_disengage(node);
		return((char)NULL);
		}

	return(char_buff);
}

static void vasenet_send_error(int errnum, 
	__uint32_t job_id, 
	__uint32_t client_fd)
{
	char resp_msg[255];

	switch (errnum)
		{
		case 1:
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:Invalid Packet",
				job_id,client_fd);
			lm_send_to_nexxus(resp_msg);
			break;
		case 2:
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:Unimplemented Command",
				job_id,client_fd);
			break;
		case 3:
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:Bad Address",
				job_id,client_fd);
			break;
		case 4:
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:Request Timed Out",
				job_id,client_fd);
			break;
		default:
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:Unknown Error Code (%d)",
				job_id,client_fd,errnum);
		}
	lm_send_to_nexxus(resp_msg);
}

static int vasenet_is_configured_master(NODELIST *node)
{
	if ((node->type == MASTER_NODE) && (node->ip_address) && (node->password))
		return(0);
	return(-1);
}
